#ifndef _SW_ATOMIC_H
#define _SW_ATOMIC_H


#ifdef  CONFIG_NUMA

#include <linux/types.h>

/*
 * Atomic operations that C can't guarantee us.  Useful for
 * resource counting etc..
 *
 * And probably incredibly slow on parisc.  OTOH, we don't
 * have to write any serious assembly.   prumpf
 */

#include <asm/spinlock.h>
#include <asm/cache.h>		/* we use L1_CACHE_BYTES */
/* Use an array of spinlocks for our atomic_ts.
 * Hash function to index into a different SPINLOCK.
 * Since "a" is usually an address, use one spinlock per cacheline.
 */
#define ATOMIC_LOCK_PAGE 0xfffffc00008c0000UL

#  define ATOMIC_HASH_SIZE 2048
//#  define ATOMIC_HASH(a) (&(__atomic_hash[ ((unsigned long) (a)) & (ATOMIC_HASH_SIZE-1) ]))
#define ATOMIC_HASH(a) (arch_spinlock_t *)(ATOMIC_LOCK_PAGE + (((unsigned long) (a)) & (ATOMIC_HASH_SIZE-1)) * sizeof(arch_spinlock_t))
extern arch_spinlock_t __atomic_hash[ATOMIC_HASH_SIZE];

/* Can't use raw_spin_lock_irq because of #include problems, so
 * this is the substitute */
#define _atomic_spin_lock_irqsave(l,f) do {	\
	arch_spinlock_t *s = ATOMIC_HASH(l);		\
	local_irq_save(f);			\
	arch_spin_lock(s);			\
} while(0)

#define _atomic_spin_unlock_irqrestore(l,f) do {	\
	arch_spinlock_t *s = ATOMIC_HASH(l);			\
	arch_spin_unlock(s);				\
	local_irq_restore(f);				\
} while(0)


/*
 * Note that we need not lock read accesses - aligned word writes/reads
 * are atomic, so a reader never sees inconsistent values.
 */

/* It's possible to reduce all atomic operations to either
 * __atomic_add_return, atomic_set and atomic_read (the latter
 * is there only for consistency).
 */

static __inline__ int __atomic_add_return(int i, atomic_t *v)
{
	int ret;
	unsigned long flags;
	_atomic_spin_lock_irqsave(v, flags);

	ret = (v->counter += i);

	_atomic_spin_unlock_irqrestore(v, flags);
	return ret;
}

static __inline__ void atomic_set(atomic_t *v, int i) 
{
	unsigned long flags;
	_atomic_spin_lock_irqsave(v, flags);

	v->counter = i;

	_atomic_spin_unlock_irqrestore(v, flags);
}

static __inline__ int atomic_read(const atomic_t *v)
{
	return (*(volatile int *)&(v)->counter);
}

#include <asm/cmpxchg.h>
/* exported interface */
#define atomic_cmpxchg(v, o, n) (cmpxchg(&((v)->counter), (o), (n)))
#define atomic_xchg(v, new) (xchg(&((v)->counter), new))

/**
 * __atomic_add_unless - add unless the number is a given value
 * @v: pointer of type atomic_t
 * @a: the amount to add to v...
 * @u: ...unless v is equal to u.
 *
 * Atomically adds @a to @v, so long as it was not @u.
 * Returns the old value of @v.
 */
static __inline__ int __atomic_add_unless(atomic_t *v, int a, int u)
{
	int c, old;
	c = atomic_read(v);
	for (;;) {
		if (unlikely(c == (u)))
			break;
		old = atomic_cmpxchg((v), c, c + (a));
		if (likely(old == c))
			break;
		c = old;
	}
	return c;
}


#define atomic_add(i,v)	((void)(__atomic_add_return( (i),(v))))
#define atomic_sub(i,v)	((void)(__atomic_add_return(-(i),(v))))
#define atomic_inc(v)	((void)(__atomic_add_return(   1,(v))))
#define atomic_dec(v)	((void)(__atomic_add_return(  -1,(v))))

#define atomic_add_return(i,v)	(__atomic_add_return( (i),(v)))
#define atomic_sub_return(i,v)	(__atomic_add_return(-(i),(v)))
#define atomic_inc_return(v)	(__atomic_add_return(   1,(v)))
#define atomic_dec_return(v)	(__atomic_add_return(  -1,(v)))

#define atomic_add_negative(a, v)	(atomic_add_return((a), (v)) < 0)

/*
 * atomic_inc_and_test - increment and test
 * @v: pointer of type atomic_t
 *
 * Atomically increments @v by 1
 * and returns true if the result is zero, or false for all
 * other cases.
 */
#define atomic_inc_and_test(v) (atomic_inc_return(v) == 0)

#define atomic_dec_and_test(v)	(atomic_dec_return(v) == 0)

#define atomic_sub_and_test(i,v)	(atomic_sub_return((i),(v)) == 0)

#define ATOMIC_INIT(i)		{ (i) }
#define ATOMIC64_INIT(i)	{ (i) }

static __inline__ s64
__atomic64_add_return(s64 i, atomic64_t *v)
{
	s64 ret;
	unsigned long flags;
	_atomic_spin_lock_irqsave(v, flags);

	ret = (v->counter += i);

	_atomic_spin_unlock_irqrestore(v, flags);
	return ret;
}

static __inline__ void
atomic64_set(atomic64_t *v, s64 i)
{
	unsigned long flags;
	_atomic_spin_lock_irqsave(v, flags);

	v->counter = i;

	_atomic_spin_unlock_irqrestore(v, flags);
}

static __inline__ s64
atomic64_read(const atomic64_t *v)
{
	return (*(volatile long *)&(v)->counter);
}

#define atomic64_add(i,v)	((void)(__atomic64_add_return( ((s64)(i)),(v))))
#define atomic64_sub(i,v)	((void)(__atomic64_add_return(-((s64)(i)),(v))))
#define atomic64_inc(v)		((void)(__atomic64_add_return(   1,(v))))
#define atomic64_dec(v)		((void)(__atomic64_add_return(  -1,(v))))

#define atomic64_add_return(i,v)	(__atomic64_add_return( ((s64)(i)),(v)))
#define atomic64_sub_return(i,v)	(__atomic64_add_return(-((s64)(i)),(v)))
#define atomic64_inc_return(v)		(__atomic64_add_return(   1,(v)))
#define atomic64_dec_return(v)		(__atomic64_add_return(  -1,(v)))

#define atomic64_add_negative(a, v)	(atomic64_add_return((a), (v)) < 0)

#define atomic64_inc_and_test(v) 	(atomic64_inc_return(v) == 0)
#define atomic64_dec_and_test(v)	(atomic64_dec_return(v) == 0)
#define atomic64_sub_and_test(i,v)	(atomic64_sub_return((i),(v)) == 0)

/* exported interface */
#define atomic64_cmpxchg(v, o, n) \
	((__typeof__((v)->counter))cmpxchg(&((v)->counter), (o), (n)))
#define atomic64_xchg(v, new) (xchg(&((v)->counter), new))

/**
 * atomic64_add_unless - add unless the number is a given value
 * @v: pointer of type atomic64_t
 * @a: the amount to add to v...
 * @u: ...unless v is equal to u.
 *
 * Atomically adds @a to @v, so long as it was not @u.
 * Returns the old value of @v.
 */
static __inline__ int atomic64_add_unless(atomic64_t *v, long a, long u)
{
	long c, old;
	c = atomic64_read(v);
	for (;;) {
		if (unlikely(c == (u)))
			break;
		old = atomic64_cmpxchg((v), c, c + (a));
		if (likely(old == c))
			break;
		c = old;
	}
	return c != (u);
}

#define atomic64_inc_not_zero(v) atomic64_add_unless((v), 1, 0)

#else      /* -----  not CONFIG_NUMA  ----- */

#include <linux/types.h>
#include <asm/barrier.h>
#include <asm/cmpxchg.h>

/*
 * Atomic operations that C can't guarantee us.  Useful for
 * resource counting etc...
 *
 * But use these as seldom as possible since they are much slower
 * than regular operations.
 */


#define ATOMIC_INIT(i)		{ (i) }
#define ATOMIC64_INIT(i)	{ (i) }

#define atomic_read(v)		READ_ONCE((v)->counter)
#define atomic64_read(v)	READ_ONCE((v)->counter)

#define atomic_set(v,i)		WRITE_ONCE((v)->counter, (i))
#define atomic64_set(v,i)	WRITE_ONCE((v)->counter, (i))

/*
 * To get proper branch prediction for the main line, we must branch
 * forward to code at the end of this object's .text section, then
 * branch back to restart the operation.
 */

static __inline__ void atomic_add(int i, atomic_t * v)
{
	unsigned long temp1, temp2, addr;
	__asm__ __volatile__(
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	"	ldi	%3, %2\n"
	"1:	lldw 	%0, 0(%3)\n"
	"	ldi 	%1, 1\n"
	"	wr_f 	%1\n"
	"	addw	%0, %4, %0\n"
	"	lstw	%0, 0(%3)\n"
	"	rd_f 	%0\n"
	"	beq 	%0, 2f\n"
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	".subsection 2\n"
	"2:	br 1b\n"
	".previous"
	:"=&r" (temp1), "=&r" (temp2), "=m" (v->counter), "=&r" (addr)
	:"Ir" (i), "m" (v->counter));
}

static __inline__ void atomic64_add(long i, atomic64_t * v)
{
	unsigned long temp1, temp2, addr;
	__asm__ __volatile__(
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	"	ldi	%3, %2\n"
	"1:	lldl 	%0, 0(%3)\n"
	"	ldi 	%1, 1\n"
	"	wr_f 	%1\n"
	"	addl	%0, %4, %0\n"
	"	lstl	%0, 0(%3)\n"
	"	rd_f 	%0\n"
	"	beq 	%0, 2f\n"
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	".subsection 2\n"
	"2:	br 1b\n"
	".previous"
	:"=&r" (temp1), "=&r" (temp2), "=m" (v->counter), "=&r" (addr)
	:"Ir" (i), "m" (v->counter));
}

static __inline__ void atomic_sub(int i, atomic_t * v)
{
	unsigned long temp1, temp2, addr;
	__asm__ __volatile__(
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	"	ldi	%3, %2\n"
	"1:	lldw 	%0, 0(%3)\n"
	"	ldi 	%1, 1\n"
	"	wr_f 	%1\n"
	"	subw	%0, %4, %0\n"
	"	lstw	%0, 0(%3)\n"
	"	rd_f 	%0\n"
	"	beq 	%0, 2f\n"
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	".subsection 2\n"
	"2:	br 1b\n"
	".previous"
	:"=&r" (temp1), "=&r" (temp2), "=m" (v->counter), "=&r" (addr)
	:"Ir" (i), "m" (v->counter));
}

static __inline__ void atomic64_sub(long i, atomic64_t * v)
{
	unsigned long temp1, temp2, addr;
	__asm__ __volatile__(
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	"	ldi	%3, %2\n"
	"1:	lldl 	%0, 0(%3)\n"
	"	ldi 	%1, 1\n"
	"	wr_f 	%1\n"
	"	subl	%0, %4, %0\n"
	"	lstl	%0, 0(%3)\n"
	"	rd_f 	%0\n"
	"	beq 	%0, 2f\n"
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	".subsection 2\n"
	"2:	br 1b\n"
	".previous"
	:"=&r" (temp1), "=&r" (temp2), "=m" (v->counter), "=&r" (addr)
	:"Ir" (i), "m" (v->counter));
}


/*
 * Same as above, but return the result value
 */
static inline int atomic_add_return(int i, atomic_t *v)
{
	long temp, result, addr;
	smp_mb();
	__asm__ __volatile__(
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	"	ldi	%3, %1\n"
	"1:	lldw	%0, 0(%3)\n"
	"	ldi	%2, 1\n"
	"	wr_f	%2\n"
	"	addw	%0, %4, %2\n"
	"	addw	%0, %4, %0\n"
	"	lstw	%0, 0(%3)\n"
	"	rd_f	%0\n"
	"	beq	%0, 2f\n"
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	".subsection 2\n"
	"2:	br 1b\n"
	".previous"
	:"=&r" (temp), "=m" (v->counter), "=&r" (result), "=&r" (addr)
	:"Ir" (i), "m" (v->counter) : "memory");
	smp_mb();
	return result;
}

static __inline__ long atomic64_add_return(long i, atomic64_t * v)
{
	long temp, result, addr;
	__asm__ __volatile__(
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	"	ldi	%3, %1\n"
	"1:	lldl	%0, 0(%3)\n"
	"	ldi	%2, 1\n"
	"	wr_f	%2\n"
	"	addl	%0, %4, %2\n"
	"	addl	%0, %4, %0\n"
	"	lstl	%0, 0(%3)\n"
	"	rd_f	%0\n"
	"	beq	%0, 2f\n"
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	".subsection 2\n"
	"2:	br 1b\n"
	".previous"
	:"=&r" (temp), "=m" (v->counter), "=&r" (result), "=&r" (addr)
	:"Ir" (i), "m" (v->counter) : "memory");
	smp_mb();
	return result;
}

static __inline__ int atomic_sub_return(int i, atomic_t * v)
{
	long temp, result, addr;
	__asm__ __volatile__(
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	"	ldi	%3, %1\n"
	"1:	lldw	%0, 0(%3)\n"
	"	ldi	%2, 1\n"
	"	wr_f	%2\n"
	"	subw	%0, %4, %2\n"
	"	subw	%0, %4, %0\n"
	"	lstw	%0, 0(%3)\n"
	"	rd_f	%0\n"
	"	beq	%0, 2f\n"
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	".subsection 2\n"
        "2:     rtc     %0\n"   
        "       and     %0,0x1f,%0\n"
        "3:     beq %0,1b\n"
        "       subl %0,1,%0\n" 
        "       br  3b\n"	
	".previous"
	:"=&r" (temp), "=m" (v->counter), "=&r" (result), "=&r" (addr)
	:"Ir" (i), "m" (v->counter) : "memory");
	smp_mb();
	return result;
}

static __inline__ long atomic64_sub_return(long i, atomic64_t * v)
{
	long temp, result, addr;
	__asm__ __volatile__(
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	"	ldi	%3, %1\n"
	"1:	lldl	%0, 0(%3)\n"
	"	ldi	%2, 1\n"
	"	wr_f	%2\n"
	"	subl	%0, %4, %2\n"
	"	subl	%0, %4, %0\n"
	"	lstl	%0, 0(%3)\n"
	"	rd_f	%0\n"
	"	beq	%0, 2f\n"
#ifdef CONFIG_SMP
	"       memb\n"
#endif
	".subsection 2\n"
	"2:	br 1b\n"
	".previous"
	:"=&r" (temp), "=m" (v->counter), "=&r" (result), "=&r" (addr)
	:"Ir" (i), "m" (v->counter) : "memory");
	smp_mb();
	return result;
}

#define atomic64_cmpxchg(v, old, new) (cmpxchg(&((v)->counter), old, new))
#define atomic64_xchg(v, new) (xchg(&((v)->counter), new))

#define atomic_cmpxchg(v, old, new) (cmpxchg(&((v)->counter), old, new))
#define atomic_xchg(v, new) (xchg(&((v)->counter), new))



/**
 * __atomic_add_unless - add unless the number is a given value
 * @v: pointer of type atomic_t
 * @a: the amount to add to v...
 * @u: ...unless v is equal to u.
 *
 * Atomically adds @a to @v, so long as it was not @u.
 * Returns the old value of @v.
 */
static __inline__ int __atomic_add_unless(atomic_t *v, int a, int u)
{
	int c, old;
	c = atomic_read(v);
	for (;;) {
		if (unlikely(c == (u)))
			break;
		old = atomic_cmpxchg((v), c, c + (a));
		if (likely(old == c))
			break;
		c = old;
	}
	return c;
}


/**
 * atomic64_add_unless - add unless the number is a given value
 * @v: pointer of type atomic64_t
 * @a: the amount to add to v...
 * @u: ...unless v is equal to u.
 *
 * Atomically adds @a to @v, so long as it was not @u.
 * Returns the old value of @v.
 */
static __inline__ int atomic64_add_unless(atomic64_t *v, long a, long u)
{
	long c, old;
	c = atomic64_read(v);
	for (;;) {
		if (unlikely(c == (u)))
			break;
		old = atomic64_cmpxchg((v), c, c + (a));
		if (likely(old == c))
			break;
		c = old;
	}
	return c != (u);
}

#ifdef CONFIG_SMP
#define MEMB         "       memb\n"
#else
#define MEMB
#endif

#define ATOMIC_OP(op, asm_op) 						\
static __inline__ void atomic_##op(int i, atomic_t * v)			\
{									\
        unsigned long temp1, temp2, addr;				\
        __asm__ __volatile__(						\
        MEMB								\
        "       ldi     %3, %2\n"					\
        "1:     lldw    %0, 0(%3)\n"					\
        "       ldi     %1, 1\n"					\
        "       wr_f    %1\n"						\
        "       " #asm_op " %0, %4, %0\n"				\
        "       lstw    %0, 0(%3)\n"					\
        "       rd_f    %0\n"						\
        "       beq     %0, 2f\n"					\
        MEMB							\
        ".subsection 2\n"						\
        "2:     br 1b\n"						\
        ".previous"							\
        :"=&r" (temp1), "=&r" (temp2), "=m" (v->counter), "=&r" (addr)  \
        :"Ir" (i), "m" (v->counter));					\
}									\



#define ATOMIC64_OP(op, asm_op)                                         \
static __inline__ void atomic64_##op(int i, atomic64_t * v)             \
{									\
        unsigned long temp1, temp2, addr;				\
        __asm__ __volatile__(						\
        "       memb\n"							\
        MEMB								\
        "       ldi     %3, %2\n"					\
        "1:     lldl    %0, 0(%3)\n"					\
        "       ldi     %1, 1\n"					\
        "       wr_f    %1\n"						\
        "       " #asm_op " %0, %4, %0\n"				\
        "       lstl    %0, 0(%3)\n"					\
        "       rd_f    %0\n"						\
        "       beq     %0, 2f\n"					\
        MEMB							\
        ".subsection 2\n"						\
        "2:     br 1b\n"						\
        ".previous"							\
        :"=&r" (temp1), "=&r" (temp2), "=m" (v->counter), "=&r" (addr)	\
        :"Ir" (i), "m" (v->counter));					\
                                                                        \
}									\

ATOMIC_OP(and, and)
ATOMIC_OP(andnot, bic)
ATOMIC_OP(or, bis)
ATOMIC_OP(xor, xor)
ATOMIC64_OP(and, and)
ATOMIC64_OP(andnot, bic)
ATOMIC64_OP(or, bis)
ATOMIC64_OP(xor, xor)


#define atomic_andnot atomic_andnot
#define atomic64_andnot atomic64_andnot

#define atomic64_inc_not_zero(v) atomic64_add_unless((v), 1, 0)

#define atomic_add_negative(a, v) (atomic_add_return((a), (v)) < 0)
#define atomic64_add_negative(a, v) (atomic64_add_return((a), (v)) < 0)

#define atomic_dec_return(v) atomic_sub_return(1,(v))
#define atomic64_dec_return(v) atomic64_sub_return(1,(v))

#define atomic_inc_return(v) atomic_add_return(1,(v))
#define atomic64_inc_return(v) atomic64_add_return(1,(v))

#define atomic_sub_and_test(i,v) (atomic_sub_return((i), (v)) == 0)
#define atomic64_sub_and_test(i,v) (atomic64_sub_return((i), (v)) == 0)

#define atomic_inc_and_test(v) (atomic_add_return(1, (v)) == 0)
#define atomic64_inc_and_test(v) (atomic64_add_return(1, (v)) == 0)

#define atomic_dec_and_test(v) (atomic_sub_return(1, (v)) == 0)
#define atomic64_dec_and_test(v) (atomic64_sub_return(1, (v)) == 0)

#define atomic_inc(v) atomic_add(1,(v))
#define atomic64_inc(v) atomic64_add(1,(v))

#define atomic_dec(v) atomic_sub(1,(v))
#define atomic64_dec(v) atomic64_sub(1,(v))

#endif     /* -----  not CONFIG_NUMA  ----- */

#endif /* _SW_ATOMIC_H */
